#include <stdio.h>
#include <X11/extensions/shape.h>
#include <pthread.h>

#include "constants.h"
#include "qlbar.h"

/**
 * ctor creates bar and fill members with default values
 */
QLBar::QLBar ()
{
	active = false;
	if (pthread_create(&hidding_thread, NULL, hideBar, this) != 0)
		fprintf(stderr, "Could not create thread\n");

	if (pthread_mutex_init(&graph_mutex, NULL) != 0)
		fprintf(stderr, "Could not init mutex");

	m_icon_width = 32;
	m_icon_height = 32;
	m_bar_x = 100;
	m_bar_y = 0;
	m_bar_time = 1;
}

/**
 * dtor frees imlib images
 */
QLBar::~QLBar()
{
	pthread_mutex_destroy(&graph_mutex);
}

/**
 * Function prepares display for main window and common items windows.
 * @return int 0 on success, -1 on failure
 */
int QLBar::Prepare()
{

	XColor fg;

	display = XOpenDisplay(getenv("DISPLAY"));

	if (!display) {	
		fprintf(stderr, "Could not open display");
		return -1;
	}

	m_defScreen = DefaultScreen(display);
	visual = DefaultVisual(display, m_defScreen);
	if (!visual)
		return -1;

	root = RootWindow(display, m_defScreen);

	colormap = DefaultColormap(display, m_defScreen);

	XColor dummy;

   XAllocNamedColor(display, colormap, "#ffffff", &fg, &dummy);

	gv.foreground = fg.pixel;
   gv.line_width = 1;
   gv.line_style = LineSolid;
	gv.function = GXcopy;

   gc = XCreateGC(display, root, GCFunction|GCForeground, &gv);
	
	ComputeXY(m_bar_x, m_bar_y, m_icon_start_x, m_icon_start_y);

	XSetWindowAttributes attrib;
	attrib.override_redirect = True;
	attrib.background_pixel = 80;

	window = XCreateWindow(display, DefaultRootWindow(display), 
		m_bar_x, m_bar_y, getBarWidth(), getBarHeight(), 0, CopyFromParent, 
		InputOutput, CopyFromParent, CWOverrideRedirect | CWBackPixel, &attrib);

   XSelectInput(display, window,
             EnterWindowMask|LeaveWindowMask|ExposureMask|ButtonPressMask|ButtonReleaseMask|ButtonMotionMask);
   XMapWindow(display, window);

	XSync(display, False);

	imlib_context_set_display(display);
	imlib_context_set_visual(visual);
	imlib_context_set_colormap(colormap);
	imlib_context_set_drawable(window);
	imlib_context_set_dither(1);
	imlib_context_set_blend(1);

	prepareItems();	

	return 0;
}


/**
 * Function returns width of a bar
 * @return int width of a bar in pixels
 */
int QLBar::getBarWidth() const
{
	if (m_pcfg->GetBarWidth() == -1)
		return (m_pcfg->GetLayout() == BAR_HORIZONTAL) ?
			items.size() * m_pcfg->GetIconWidth() : 
			m_pcfg->GetIconHeight();

	else
		return m_pcfg->GetBarWidth();
}

/**
 * Function returns height of a qlbar
 * @return int height of a bar in pixels
 */
int QLBar::getBarHeight() const
{
	if (m_pcfg->GetBarHeight() == -1)
		return (m_pcfg->GetLayout() == BAR_HORIZONTAL) ? 
			items.size() * m_pcfg->GetIconHeight() : 
			m_pcfg->GetIconWidth();

	else
		return m_pcfg->GetBarHeight();
}


/**
 * Function prepare UI data for every item on a list
 */
void QLBar::prepareItems()
{
	int icon_x = m_icon_start_x;
	int icon_y = m_icon_start_y;
	vector<QLItem *>::iterator it = items.begin();
	for ( ; it != items.end(); it++){
		(*it)->CreateUI(display, window, icon_x, icon_y, m_pcfg->GetIconWidth(), m_pcfg->GetIconHeight());
		switch(m_pcfg->GetLayout()){
			case BAR_HORIZONTAL:
				icon_x += m_pcfg->GetIconWidth();
				break;
			case BAR_VERTICAL:
				icon_y += m_pcfg->GetIconHeight();
				break;
		}
	}
}


/**
 * Function receive events from created windows
 * return bool true on success, false on failure
 */
int QLBar::Run()
{
	hideAllItems();

	XEvent ev;
	while (1) {
		XNextEvent(display, &ev);
		DBGOUT printf("Event: %d\n", ev.type);
		switch (ev.type){
			case Expose:
				{
				DBGOUT printf("expose\n");
				QLItem * item = findItem(ev.xexpose.window);
	
				if (pthread_mutex_lock(&graph_mutex) != 0){
					printf("dupa");
				}

				if (item != NULL)
					item->Paint();

		      XClearWindow(display, window);

				pthread_mutex_unlock(&graph_mutex);
				break;
				}
			case LeaveNotify:
				{
					DBGOUT printf("leave\n");
					//QLItem * item = findItem(ev.xcrossing.window);
					//if (item != NULL)
					//	item->Hide(); unblink item
					active = true;
					pthread_create(&hidding_thread, NULL, hideBar, this);
				}
				break;
			case EnterNotify:
				DBGOUT printf("enter\n");
				if (ev.xcrossing.window == window)
				{
					//TODO: show animated
					pthread_cancel(hidding_thread);
//					ScaleItems(m_icon_height);
					showAllItems();
				}
				else
				{
		//			active = findItem(ev.xcrossing.window);
					pthread_cancel(hidding_thread);
					//TODO: remember last entered
				}
				break;
			case ButtonPress:
				{
					QLItem * item = findItem(ev.xcrossing.window);
					if (item!=NULL)
						item->Execute();
				}
				break;
			default:
				break;
		}
	};
return 0;
}


/**
 * Function iterates through vector containing qlitems and compares their
 * window property.
 * @param window pattern window
 * @return QLItem pointer to the item found, NULL when not found
 */
QLItem * QLBar::findItem(const Window window) const
{
	vector<QLItem *>::const_iterator it = items.begin();

	for (;it != items.end(); it++)
	{
		if ( (*it)->GetWindow() == window)
			return (*it);
	}
	return NULL;
}

/**
 * Function iterates through vector containing qlitems and hide each of them
 */
void QLBar::hideAllItems() 
{
	pthread_mutex_lock(&graph_mutex);

	vector<QLItem *>::const_iterator it = items.begin();
	for (; it != items.end() ; it++)
		(*it)->Hide();

	XSync(display, False);

	pthread_mutex_unlock(&graph_mutex);
}

/**
 * Function iterates through vector containing qlitems and show each of them
 */
void QLBar::showAllItems()
{
	pthread_mutex_lock(&graph_mutex);

	vector<QLItem *>::const_iterator it = items.begin();
	for (; it != items.end() ; it++)
		(*it)->Show();
	XSync(display, False);

	pthread_mutex_unlock(&graph_mutex);
}


/**
 * Function appends item onto the end of a list
 * @param pItem pointer to newly created item
 * @return int as size of a vector containing items
 */
int QLBar::AddItem ( QLItem * pItem )
{
	items.push_back(pItem);
	return items.size();
}


/**
 * Function inserts item at a given index
 * @param index integer index of inserting element
 * @param pItem pointer to a newly created item
 * @return int as size of a vector containing items
 */
int QLBar::AddItem ( const int & index, QLItem * pItem )
{
	int i = 0;
	vector<QLItem *>::iterator it = items.begin();
	for ( ; it != items.end(); it++,i++){
		if (i==index)
			break;
	}

	if (it != items.begin())
		items.insert(it, pItem);
	else 
		items.push_back(pItem);

	return items.size();
}

/**
 * Function computes coordinates of bar.
 * Coordinates depends on settings: position, layout, and screen width and height
 * @param x reference to x where x computed coordinate will be placed
 * @param y reference to y where y copmuted coordinate will be placed
 */
void QLBar::ComputeXY(int & x, int & y, int & ic_x, int & ic_y) 
{
	int bw = m_pcfg->GetBarWidth();
	int bh = m_pcfg->GetBarHeight();
	int dw = DisplayWidth(display, m_defScreen);
	int dh = DisplayHeight(display, m_defScreen);
	int ih = m_pcfg->GetIconHeight();
	int iw = m_pcfg->GetIconWidth();
	int layout = m_pcfg->GetLayout();
	int ho = m_pcfg->GetHOffset();
	int vo = m_pcfg->GetVOffset();

	switch(m_pcfg->GetPosition()){
		case BP_NORTH:
			if (layout == BAR_HORIZONTAL){
				y = ic_y = 0;
				x = ic_x = dw / 2 - bw / 2;
			}
			else {
				// not available
			}
			break;
		case BP_SOUTH:
			if (layout == BAR_HORIZONTAL){
				y = dh - bw;
				x = ic_x = dw / 2 - bw / 2;
				ic_y = y - ih;
			}
			else {
				// not available
			}
			break;
		case BP_NORTH_EAST:
			y = ic_y = 0;
			x = dw - bw;
			if (layout == BAR_HORIZONTAL){
				x -= ho;
				ic_x = x;
			}
			else {
				y += vo;
				ic_x = dw - iw;
			}
			break;
		case BP_EAST:
			if (layout == BAR_VERTICAL){
				x = dw - bw;
				y = ic_y = dh / 2 - bh / 2;
				ic_x = dw - iw;
			}
			else {
				// not available
			}
			break;
		case BP_SOUTH_EAST:
			x = dw - bw; 
			if (layout == BAR_HORIZONTAL){
				x -= ho;
				y = dh - bw;
				ic_x = x;
				ic_y = y - ih;
			}
			else{
				y = dh - bh - ho;
				ic_x = dw - iw;
				ic_y = dh - bh;
			}
			break;
		case BP_NORTH_WEST:
			if (layout == BAR_HORIZONTAL){
				y = ic_y = 0;
				x = ic_x = ho;
			}
			else {
				y = ic_y = vo;
				x = ic_x = 0;
			}
			break;
		case BP_WEST:
			if (layout == BAR_VERTICAL){
				x = ic_x = 0;
				y = ic_y = dh / 2 - bh / 2;
			}
			else {
				// not available
			}
			break;
		case BP_SOUTH_WEST:
			y = dh;
			if (layout == BAR_HORIZONTAL){
				x = ic_x = ho;	  
				ic_y = y - ih;
			}
			else {
				x = ic_x = 0;
				ic_y = dw - bh - vo;
			}
			break;
	}
}


void QLBar::ScaleItems()
{
	vector<QLItem *>::iterator it = items.begin();
	for ( ; it != items.end(); it++){
		(*it)->Scale(m_icon_width, m_icon_height);
	}
}

int QLBar::GetBarX() const 
{
	return m_bar_x;
}

void QLBar::SetDock(const int &pos)
{
	m_dock = pos;
}

void QLBar::SetBarTime(const int &mili)
{
	m_bar_time = mili;
}

int QLBar::GetBarTime() const
{
	return m_pcfg->GetBarTime();
}

void QLBar::SetIconDimension(const int &width, const int &height, const bool &redraw)
{
	m_icon_width = width;
	m_icon_height = height;
//	if (redraw)
}

void QLBar::SetBarColor(const int &color)
{
	m_bar_color = color;
}

void QLBar::SetBarDimension(const int &bar_width, const int &bar_height, const bool &redraw)
{
	m_bar_width = bar_width;
	m_bar_height = bar_height;
// if (redraw)
}

bool QLBar::GetActive() const 
{
	return active;
}

void QLBar::SetConfig(QLConf * cfg)
{
	if (cfg == NULL)
		return;

	m_pcfg = cfg;
	// and create valid items lise
	vector<QLItem*> * pV = cfg->GetItems();
	vector<QLItem*>::iterator it = pV->begin();
	for (; it!=pV->end(); it++){
		if (!(*it)->isFailed()){ // destroy unnessesary icons
			AddItem(*it);
		}
	}
}


////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
// thread function //
void* hideBar(void * thid)
{
	QLBar * qb = static_cast<QLBar *> (thid);
	if (!qb->GetActive())
		pthread_exit(0);

	if (qb == NULL) {
		pthread_exit(0);
	}

	sleep(qb->GetBarTime()); // wait a second and then hide all items
	if (qb!=NULL)
		qb->hideAllItems();
	

	pthread_exit(0);
}
