#include <regex.h>
#include <fstream>
#include <vector>

#include "qlconf.h"
#include "constants.h"





/**
 * ctor
 * fills members with default values;
 */
QLConf::QLConf()
{
	b_icon_width = false;
	b_icon_height = false;
	b_bar_height = false;
	b_bar_width = false;
	b_bar_color = false;
	b_position = false;
	b_layout = false;
	b_bar_time = false;
	b_hoffset = false;
	b_voffset = false;

   icon_width = 32;
   icon_height = 32;
   bar_color = 80;
   bar_width = -1;
   bar_height = 1;
   position = BP_NORTH; // top
	layout = BAR_HORIZONTAL;
   bar_time = 1;
	hoffset = 0;
	voffset = 0;
}

/**
 * dtor
 * clears items list
 */
QLConf::~QLConf()
{
	for (unsigned int i=0;i<items.size();i++) {
		if (items[i] != NULL) {
			delete items[i];
			items[i] = NULL;
		}
	}
}

/**
 * Function compares two strings
 * @param opt pointer to first string
 * @param pattern pointer to second string used as pattern
 * @return 0 when strings are the same
 * @todo make comparision not case sensitive
 */
int QLConf::comp(const char* opt, const char* pattern) const {
	if (strlen (opt) == strlen(pattern))
		return strncmp(opt,pattern, strlen(pattern));
	else 
		return -1;
}



/**
 * Function copies src value into dst pointer
 * @param dst pointer to integer destination
 * @param src source integer value
 * @param state pointer to state that will be modified after copy
 * @param ov boolean that say the destination value should be overriden
 */
void QLConf::Set(int*dst,const int &src, bool *state, const bool ov)
{
	if ((ov == true) || (*state == false)){
		*dst = src;
		*state = true;
	}
}



/**
 * Function compares given option name with well known patterns and
 * when option name is valid config member is set with given value.
 * @param optname pointer name of the option to set
 * @param value pointer to string representation of the value to set
 * @param ov destination override flag, true means that value will be overriden
 * @return int 0 on success, -1 on failure eg. option name is not known
 */
int QLConf::Parse(const char * optname, const char * value, const bool ov)
{
	if (!comp(optname, "icon-width"))
		Set(&icon_width,atoi(value),&b_icon_width, ov);

	else if (!comp(optname, "icon-height"))
		Set(&icon_height, atoi(value),&b_icon_height, ov);

	else if (!comp(optname, "bar-width")){
		if (!comp(value, "default"))
			Set(&bar_width, -1, &b_bar_width, ov);
		else
			Set(&bar_width, atoi(value), &b_bar_width, ov);
	}
	else if (!comp(optname, "bar-height"))
		Set(&bar_height, atoi(value),&b_bar_height, ov);

	else if (!comp(optname, "bar-color"))
		Set(&bar_color, atoi(value),&b_bar_color, ov);

	else if (!comp(optname, "position")) {
		if (!comp(value, "n"))
			Set(&position, BP_NORTH, &b_position, ov);
		else if (!comp(value, "ne"))
			Set(&position, BP_NORTH_EAST, &b_position, ov);
		else if (!comp(value, "e"))
			Set(&position, BP_EAST, &b_position, ov);
		else if (!comp(value, "se"))
			Set(&position, BP_SOUTH_EAST, &b_position, ov);
		else if (!comp(value, "s"))
			Set(&position, BP_SOUTH, &b_position, ov);
		else if (!comp(value, "sw"))
			Set(&position, BP_SOUTH_WEST, &b_position, ov);
		else if (!comp(value, "w"))
			Set(&position, BP_WEST, &b_position, ov);
		else if (!comp(value, "nw"))
			Set(&position, BP_NORTH_WEST, &b_position, ov);
		else
		{
			fprintf(stderr, "Unknown value for option %s = %s\n", optname, value);
			return -1;
		}
	} 
	else if (!comp(optname, "layout")) {
		if (!comp(value, "vertical"))
			Set(&layout, BAR_VERTICAL, &b_layout, ov);
		else if (!comp(value, "horizontal"))
			Set(&layout, BAR_HORIZONTAL, &b_layout, ov);
	}
	else if (!comp(optname, "bar-time")){
		Set(&bar_time, atoi(value),&b_bar_time, ov);
	}
	else if (!comp(optname, "hoffset")){
		Set(&hoffset, atoi(value), &b_hoffset, ov);
		}
	else if (!comp(optname, "voffset")){
		Set(&voffset, atoi(value), &b_voffset, ov);
	}
	else
	{
		fprintf(stderr, "Unknown option: %s = %s\n", optname, value);
		return -1; // option unknown
	}	

	return 0;
}


/**
 * Function reads menu file, creates QLItems list
 * @param filename name of file to read
 * @return 0 on success, -1 on failure
 */
int QLConf::ReadMenuConfig( const char * filename )
{

	char buffer [1024];

	regex_t rexec;
	regmatch_t match[4];

	char * pattern = 
		"^[[:blank:]]*\\[exec\\][[:blank:]]*\\((.*)\\)[[:blank:]]*\\{(.*)\\}[[:blank:]]*<(.*)>";

	if(regcomp(&rexec,pattern, REG_EXTENDED ) != 0)
	{
		fprintf(stderr, "error compile regex");
		return -1;
	}

	ifstream cfg;
	cfg.open(filename, ifstream::in);
	if (!cfg){
		fprintf(stderr, "cannot open config file: %s\n", filename);
		return -1;
	}

	while (cfg.good()){
		cfg.getline(buffer,sizeof(buffer));
		if (strlen(buffer)==0) continue;
		
		if (regexec(&rexec, buffer, 4, match, REG_NOTEOL) == 0)
		{
			char ** appdesc = new char * [3];
			for (int i=0;i<3;i++)
			{
				*(appdesc+i) = new char [match[i+1].rm_eo - match[i+1].rm_so +1];
				strncpy(*(appdesc+i), buffer+match[i+1].rm_so, match[i+1].rm_eo - match[i+1].rm_so);
				(*(appdesc+i))[match[i+1].rm_eo - match[i+1].rm_so] = '\0';
			}

			//TODO: first level validate params
			
			QLItem * item = new QLItem (*(appdesc+2), *(appdesc+1), *(appdesc));
			items.push_back(item);
			
			for (int i=0;i<3;i++)
			{
				if (*(appdesc+i) != NULL)
					delete [] *(appdesc+i);
				*(appdesc+i) = NULL;
			}

			if (*(appdesc)!=NULL)
				delete [] appdesc;
			appdesc = NULL;
		}
	}
	cfg.close();	
	regfree(&rexec);
	return 0;
}


/**
 * Function reads config 
 * @param filename name of file to read
 * @return 0 on success, -1 on failure
 */
int QLConf::ReadConfigFile(const char * filename)
{

	char buffer [1024];

	regex_t rexec;
	regmatch_t match [3];

	char * pattern = "^([a-zA-Z-]+)[[:blank:]]*=[[:blank:]]([a-zA-Z0-9-]*)";

	if(regcomp(&rexec,pattern, REG_EXTENDED ) != 0) {
		fprintf(stderr, "error compile regex");
		return -1;
	}  

	ifstream cfg;
	cfg.open(filename, ifstream::in);
	if (!cfg){
		fprintf(stderr, "cannot open config file: %s\n", filename);
		return -1;
	}

	while (cfg.good()){
		cfg.getline(buffer, sizeof(buffer));
		if (strlen(buffer) == 0)
			continue;
		if (*buffer == '#') // skip comment 
			continue;

		if (regexec(&rexec, buffer, 3, match, 0) == 0){
			// matched
			char * opt_name = new char [ match[1].rm_eo - match[1].rm_so + 1];
			strncpy(opt_name, buffer + match[1].rm_so, match[1].rm_eo - match[1].rm_so);
			opt_name[match[1].rm_eo - match[1].rm_so] = '\0';

			char * value = new char [ match[2].rm_eo - match[2].rm_so + 1];
			strncpy(value, buffer + match[2].rm_so, match[2].rm_eo - match[2].rm_so);
			value[match[2].rm_eo - match[2].rm_so] = '\0';

			Parse(opt_name, value, false);

			delete [] opt_name;
			opt_name = NULL;
			delete [] value;
			value = NULL;
		}
	}

	cfg.close();
	return 0;
}


/**
 * Function returns width of the bar.
 * @param asis bar width is not computed. It's returned AS IS 
 * @return int widht of the bar
 */
int QLConf::GetBarWidth(bool asis) const
{
	if ((asis == false || bar_width == -1) && layout == BAR_HORIZONTAL)
		return icon_width * items.size();
	else 
		return bar_width;
}


/**
 * Function returns height of the bar. 
 * @param asis bar height is not computed. It's returned AS IS
 * @return int height of the bar
 */
int QLConf::GetBarHeight(bool asis) const
{
	if ((asis == false || bar_height == -1) && layout == BAR_VERTICAL)
		return icon_height * items.size();
	else
		return bar_height;
}


/**
 * Function puts string given by parameter onto standard error when
 * second parameter is true
 * @param msg message string
 * @param verbose if true message is displayed on stderr, nothing happen else
 */
void QLConf::Error(const char * msg, const bool verbose) const
{
	if (verbose)
		fprintf(stderr, msg);
}


/**
 * Function validates if config's data is correct.
 * @param verbose boolean true if error information should be displayed on stderr
 * @return bool true if config's data is valid, false on invalid
 */
bool QLConf::Validate(const bool verbose) const
{
	int valid = true;
	switch (this->GetLayout()){
		case BAR_HORIZONTAL:
			switch (this->GetPosition()){
				case BP_EAST:
				case BP_WEST:
					valid = false;
					Error("Bar position EAST or WEST cannot be set with layout HORIZONTAL\n", verbose);
					break;
				default:
					break;
			}
			if (this->GetVOffset() != 0){
				valid = false;
			 	Error("Vertical offset cannot be set with layout HORIZONTAL\n", verbose);
			}
			break;
		case BAR_VERTICAL:
			switch (this->GetPosition()){
				case BP_NORTH:
				case BP_SOUTH:
					valid = false;
					Error("Bar position NORTH or SOUTH cannot be set with layout VERTICAL\n", verbose);
					break;
				default:
					break;
			}
			if (this->GetHOffset() != 0){
				valid = false;
				Error("Horizontal offset cannot be set with layout VERTICAL\n", verbose);
			}
		break;
	}

	if (this->GetHOffset() < 0){
		valid = false;
		Error("Horizontal offset cannot be lower than 0\n", verbose);
	}
	
	if (this->GetVOffset() < 0){
		valid = false;
		Error("Vertical offset cannot be lower than 0\n", verbose);
	}

	return valid;

}

