#include <stdio.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <X11/extensions/shape.h>

#include <string>
#include "constants.h"
#include "qlitem.h"



/**
 * ctor fills members with default value, reads icons
 * and creates images
 */
QLItem::QLItem ( char * ip, char * ep, char * fn)
{
	m_width = 64;
	m_height = 64;

	failed = false;

	m_x = 0;
	m_y = 0;
	int iip = strlen (ip);
	m_pIconPath = new char [ iip + 1 ];
	strncpy(m_pIconPath, ip, iip);
	m_pIconPath[iip] = '\0';

	iip = strlen (ep);
	m_pExecPath = new char [ iip + 1 ];
	strncpy(m_pExecPath, ep, iip);
	m_pExecPath[iip] = '\0';

	iip = strlen (fn);
	m_pName = new char [ iip +1 ];
	strncpy(m_pName, fn, iip);
	m_pName[iip] = '\0';

	Init();
}

void QLItem::createDefIconPath(char * &dst)
{
	if (dst != NULL)
		delete [] dst;

	const char * home = getenv("HOME");
	int home_len = strlen(home);
  	int qlbdir_len = strlen(QLB_DEF_DIR);
	
	int len = strlen ( QLB_DEF_ICON );
	dst = new char [ home_len + qlbdir_len + len + 2 ];
	strncpy (dst, home, home_len + 1);
   strncat (dst, "/", 1);
   strncat (dst, QLB_DEF_DIR, qlbdir_len + 1);
   strncat (dst, QLB_DEF_ICON, len+1);

	
}

/**
 * Function initializes and validates item's data
 *
 */
void QLItem::Init()
{
	// check paths
	struct stat st;
	stat(m_pIconPath, &st);

	if (!S_ISREG(st.st_mode)){
		fprintf(stderr, "Could not find image: %s\nUsing default\n", m_pIconPath);
		createDefIconPath(m_pIconPath);
		}

	Imlib_Image tmp_im;
	tmp_im = imlib_load_image(m_pIconPath);
	if (tmp_im == NULL) {
		fprintf(stderr, "Could not load image: %s\nUsing default\n", m_pIconPath);
		char * dst = NULL;
		createDefIconPath(dst);
		tmp_im = imlib_load_image(dst);
		if (tmp_im == NULL) {
			fprintf(stderr, "Could load default icon file: %s\n", dst);
			failed = true;
		}

		if (dst != NULL){
			delete [] dst;
			dst = NULL;
		}
			
	}

	if (failed == true) // failed
		return;

	imlib_context_set_image(tmp_im);
	int width = imlib_image_get_width();
	int height = imlib_image_get_height();

	im = imlib_create_cropped_scaled_image(0,0,width,height,m_width,m_height);
	imlib_free_image();

}

/**
 * dtor frees images
 */
QLItem::~QLItem ()
{
	if (m_pIconPath != NULL) {
		delete [] m_pIconPath;
		m_pIconPath = NULL;
	}

	if (m_pExecPath != NULL) {
		delete [] m_pExecPath;
		m_pExecPath = NULL;
	}

	if (m_pName != NULL) {
		delete [] m_pName;
		m_pName = NULL;
	}

	if (failed == false){
		imlib_context_set_image(im);
		imlib_free_image();
	}
}

/**
 * Function paints icon on a screen
 */
void QLItem::Paint()
{
	DBGOUT printf("paint : %d, %d, %d, %d\n", m_x, m_y, m_width, m_height);
	imlib_context_set_image(im);
	imlib_context_set_drawable(m_window);
	imlib_image_set_has_alpha(1);
	imlib_image_set_irrelevant_alpha(0);

	imlib_render_pixmaps_for_whole_image_at_size(
		&pix, &mask, m_width, m_height);

	XSetWindowBackgroundPixmap(m_display, m_window, pix);
	XShapeCombineMask(m_display, m_window, ShapeBounding, 0, 0, mask, ShapeSet);

	XRaiseWindow(m_display, m_window);
	XClearWindow(m_display, m_window);
}

/**
 * Function creates Window for the item and registers events.
 * @param display parents display pointer
 * @param window parents window
 * @param x coordinate x relative to parent
 * @param y coordinate y relative to parent
 * @param w width of icon
 * @param h height of icon
 */
bool QLItem::CreateUI (Display * display, Window window, 
	const int & x, const int & y, const int &w, const int &h)
{
	DBGOUT  printf("prepare: %d, %d, %d, %d\n", x, y, w, h);
	m_x = x;
	m_y = y;
	m_width = w;
	m_height = h;

	m_display = display;

	XSetWindowAttributes attrib;
 	attrib.override_redirect = True;
   attrib.background_pixel = 1;

	// create dedicated window for item
	m_window = XCreateWindow(display, DefaultRootWindow(display),
      m_x, m_y, m_width, m_height, 0, CopyFromParent,
      InputOutput, CopyFromParent, CWOverrideRedirect | CWBackPixel, &attrib);

	Show();
	
	return true;
}

void QLItem::Hide()
{
	XSelectInput(m_display, m_window, 0);
	XUnmapWindow(m_display, m_window);
}

void QLItem::Show()
{

	XSelectInput(m_display, m_window, 
		EnterWindowMask|LeaveWindowMask|ExposureMask|ButtonPressMask|ButtonReleaseMask|ButtonMotionMask);

   XMapWindow(m_display, m_window);
}


Window QLItem::GetWindow() const 
{
	return m_window;
}

void QLItem::Scale(const int & w, const int &h)
{
	m_width = w;
	m_height = h;
}


/**
 * Function executes items application as separated
 */
void QLItem::Execute() 
{
	DBGOUT printf("Execute\n");
 	pid_t pid;

   if (!(pid = fork())) 
	{
      setsid();
      switch (fork()) 
		{
      	case 0:
        		execlp("/bin/sh", "sh", "-c", m_pExecPath, NULL);
      	default:
         	_exit(0);
      }
   }

   if (pid > 0)
      wait(NULL);

}
